import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

-- General functions
function createTarget()
	local t = Turbine.UI.Label();
	t.MaximumMorale = 0;
	t.CurrentMorale = 1;
	t.Name = "";
	t.TrackingRat = "";
	t.PercentageLabel = Turbine.UI.Label();
	t.PercentageLabel.ParentTarget = t;
	return t;
end

function AddCallback(object, event, callback)
	if ( object[event] == nil ) then
		object[event] = callback;
	else
		if ( type(object[event]) == "table" ) then
			table.insert(object[event], callback);
		else
			object[event] = {object[event], callback};
		end
	end
end

function getPercentage(target)
	local text = "-";
	if ( target.MaximumMorale ~= nil and target.MaximumMorale ~= 0 ) then 
		text = round(target.CurrentMorale / target.MaximumMorale * 100, 2) .. "%";
	end
	return text;
end

function round(num, idp)
  local mult = 10^(idp or 0);
  return math.floor(num * mult + 0.5) / mult;
end

function table.find(tbl, element)
	for idx, value in pairs(tbl) do
		if ( value == element ) then return idx; end
	end
	return 0;
end

function styleLabel(lbl, x, y, w, h, font, color, align, text)
	lbl:SetPosition(x,y);
	lbl:SetSize(w,h);
	lbl:SetFont(font);
	lbl:SetText(text);
	lbl:SetTextAlignment(align);
	lbl:SetForeColor(color);
	lbl:SetMouseVisible(false);
	lbl:SetZOrder(3);
end

-- Window Dragging functions
is_dragging = false;
function startDrag(sender, args)
	if ( args.Button ~= Turbine.UI.MouseButton.Left ) then return; end
	startX = args.X;
	startY = args.Y;
	is_dragging = true;
end

function endDrag(sender, args)
	if ( is_dragging ) then
		main_window:SetLeft(main_window:GetLeft() + (args.X - startX));
		main_window:SetTop(main_window:GetTop() + (args.Y - startY));
		is_dragging = false;
	end
end

function doDrag(sender, args)
	if ( is_dragging ) then
		main_window:SetLeft(main_window:GetLeft() + (args.X - startX));
		main_window:SetTop(main_window:GetTop() + (args.Y - startY));
	end
end

function attachDragListener(target)
	target.MouseDown = startDrag;
	target.MouseUp = endDrag;
	target.MouseMove = doDrag;
end

-- Persistance functions
function saveState()
	local obj = Turbine.Object();
	obj.window_y = main_window:GetTop();
	obj.window_x = main_window:GetLeft();
	obj.target_left = left_target.Name;
	obj.target_right = right_target.Name;
	obj.morale_diff = MORALE_DIFFERENCE;
	Turbine.PluginData.Save(Turbine.DataScope.Character, "Cee_MoraleDiff", obj, nil);
end

function loadState()
	local persist = Turbine.PluginData.Load(Turbine.DataScope.Character, "Cee_MoraleDiff", nil);
	if persist ~= nil then
		main_window:SetTop(persist.window_y);
		main_window:SetLeft(persist.window_x);
		if ( persist.morale_diff ~= nil ) then MORALE_DIFFERENCE = persist.morale_diff; end
		if ( persist.target_left ~= nil and persist.target_left ~= "" ) then left_target.Name = persist.target_left; left_target:SetText(left_target.Name); end
		if ( persist.target_right ~= nil and persist.target_right ~= "" ) then right_target.Name = persist.target_right; right_target:SetText(right_target.Name); end
	end
end

function reloadPlugin()
	Turbine.PluginManager.LoadPlugin("MoraleDiffReloader");
end

-- Event handlers
frame_counter = 0;
unloaded_reloader = false;
function onUpdate()
	if ( not unloaded_reloader ) then Turbine.PluginManager.UnloadScriptState("MoraleDiffReloader"); unloaded_reloader = true; end
	
	frame_counter = frame_counter + 1;
	if ( frame_counter == UPDATE_FRAMECOUNT ) then
		main_window:SetWantsUpdates(false);
		
		if ( scan_rats ) then scanRatTargets(); end
		updateMoraleValues(left_target);
		updateMoraleValues(right_target);
		updateDifferenceBar();
		
		frame_counter = 0;
		main_window:SetWantsUpdates(true);
	end
end

function fetchTarget(sender, args)
	if ( args ~= nil and args.Button ~= Turbine.UI.MouseButton.Left ) then return; end
	if ( my_player:GetTarget() == nil ) then
		Turbine.Shell.WriteLine("Please select the target to track.");
	else
		sender.ParentTarget.Name = my_player:GetTarget():GetName();
		sender.ParentTarget:SetText(my_player:GetTarget():GetName());
		scan_rats = true;
	end
end

function processChatMessage(sender, args)
	if ( args.ChatType == Turbine.ChatType.Death ) then
		findTargetInText(left_target, args.Message);
		findTargetInText(right_target, args.Message);
	end
	if ( args.ChatType == Turbine.ChatType.Standard ) then
		if ( findInText(" assist target", args.Message) ) then reloadPlugin(); end
	end
end

function findInText(find, text)
	local idx = string.find(text, find, 1, true);
	return ( idx ~= nil and idx > 1);
end

function findTargetInText(target, text)
	local idx = string.find(text, target.Name, 1, true);
	if ( idx ~= nil and idx > 1) then
		target.TrackingRat = "";
		target.CurrentMorale = 0;
	end
end

-- Window functions
function createMainWindow()
	main_window = Turbine.UI.Window();
	main_window:SetPosition(100,100);
	main_window:SetSize(300,40);
	main_window:SetBackground("CB/MoraleDiff/Resources/window_background.tga");
	attachDragListener(main_window);
	
	main_window.targetCounter = Turbine.UI.Label();
	main_window.targetCounter:SetParent(main_window);
	styleLabel(main_window.targetCounter, 140, 3, 20, 15, Turbine.UI.Lotro.Font.Verdana10, Turbine.UI.Color(200/255,200/255,200/255), Turbine.UI.ContentAlignment.MiddleCenter, "");
	
	styleLabel(left_target, 3, 3, 140, 15, Turbine.UI.Lotro.Font.TrajanPro15, Turbine.UI.Color(245/255,223/255,146/255), Turbine.UI.ContentAlignment.MiddleLeft, "Left Target");
	styleLabel(right_target, 156, 3, 140, 15, Turbine.UI.Lotro.Font.TrajanPro15, Turbine.UI.Color(245/255,223/255,146/255), Turbine.UI.ContentAlignment.MiddleRight, "Right Target");
	styleLabel(left_target.PercentageLabel, 3, 23, 145, 15, Turbine.UI.Lotro.Font.Verdana12, Turbine.UI.Color(1,1,1), Turbine.UI.ContentAlignment.MiddleLeft, "-");
	styleLabel(right_target.PercentageLabel, 152, 23, 144, 15, Turbine.UI.Lotro.Font.Verdana12, Turbine.UI.Color(1,1,1), Turbine.UI.ContentAlignment.MiddleRight, "-");
	
	left_target.PercentageLabel:SetMouseVisible(true);
	right_target.PercentageLabel:SetMouseVisible(true);
	left_target.PercentageLabel.MouseDoubleClick = fetchTarget;
	right_target.PercentageLabel.MouseDoubleClick = fetchTarget;
	attachDragListener(left_target.PercentageLabel);
	attachDragListener(right_target.PercentageLabel);
	
	left_target:SetParent(main_window);
	left_target.PercentageLabel:SetParent(main_window);
	right_target:SetParent(main_window);
	right_target.PercentageLabel:SetParent(main_window);
	
    context_menu = Turbine.UI.ContextMenu();
    local items = context_menu:GetItems();
    items:Add(Turbine.UI.MenuItem("Set left target"));
    items:Add(Turbine.UI.MenuItem("Set right target"));
    items:Add(Turbine.UI.MenuItem("Settings"));
    
    items:Get(1).Click = function(sender, args) fetchTarget(left_target.PercentageLabel, nil); end
    items:Get(2).Click = function(sender, args) fetchTarget(right_target.PercentageLabel, nil); end
    items:Get(3).Click = function(sender, args) showSettingsWindow(); end

    main_window.MouseClick = showContextMenu;
    left_target.PercentageLabel.MouseClick = showContextMenu;
    right_target.PercentageLabel.MouseClick = showContextMenu;
	
end

function createSettingsWindow()
	settings_window = Turbine.UI.Lotro.Window();
	settings_window:SetText("Settings");
	settings_window:SetSize(350, 150);
	settings_window:SetPosition((Turbine.UI.Display.GetWidth() - 350)/2,(Turbine.UI.Display.GetHeight() - 150)/2);
	settings_window:SetZOrder(100);
	
	local lbl = Turbine.UI.Label();
	lbl:SetParent(settings_window);
	styleLabel(lbl, 20, 50, 200, 15, Turbine.UI.Lotro.Font.Verdana14, Turbine.UI.Color(1,1,1), Turbine.UI.ContentAlignment.MiddleLeft, "Maximum morale difference: ");
	
	settings_window.MaxMoraleDiff = Turbine.UI.Lotro.TextBox();
	local txt = settings_window.MaxMoraleDiff;
	txt:SetParent(settings_window);
	txt:SetSize(100, 20);
	txt:SetPosition(210, 48);
	txt:SetMultiline(false);
	txt:SetFont(Turbine.UI.Lotro.Font.Verdana12);
	txt:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
	txt:SetText(MORALE_DIFFERENCE);
	
	local btn = Turbine.UI.Lotro.Button();
	btn:SetParent(settings_window);
	btn:SetSize(60,20);
	btn:SetPosition(145,90);
	btn:SetText("OK");
	btn.MouseClick = applySettings;
end

function showSettingsWindow()
	settings_window.MaxMoraleDiff:SetText(MORALE_DIFFERENCE);
	settings_window:SetVisible(true);
	settings_window.Focus();
end

function applySettings(sender, args)
	if ( args.Button == Turbine.UI.MouseButton.Left ) then
		local newVal = tonumber(settings_window.MaxMoraleDiff:GetText());
		if (newVal ~= nil and math.floor(newVal) > 0) then MORALE_DIFFERENCE = math.floor(newVal); end
		settings_window:Close();
	end
end

function showContextMenu(sender, args)
	if ( args.Button == Turbine.UI.MouseButton.Right ) then
		context_menu:ShowMenu();
	end
end

function createDifferenceBar()
	local bar = Turbine.UI.Label();
	bar:SetMouseVisible(false);
	bar:SetBackground("CB/MoraleDiff/Resources/bar_fill.tga");
	bar:SetPosition(152,24);
	bar:SetSize(0,14);
	bar:SetParent(main_window);
	bar:SetZOrder(1);
	return bar;
end

-- Tracking functions
function targetChanged(sender, args)
	local t = sender:GetTarget();
	rat_targets[sender:GetName()] = t;
	scan_rats = true;
end

function analyzeRatTarget(rat, target, tracker, found)
	if ( target:GetName() == tracker.Name and ( rat == my_player:GetName() or not found ) ) then
		tracker.TrackingRat = rat;
		return true;
	end
	return found;
end

function scanRatTargets()
	local foundLeft = false;
	local foundRight = false;
	
	for rat, target in pairs(rat_targets) do
		foundLeft = analyzeRatTarget(rat, target, left_target, foundLeft);
		foundRight = analyzeRatTarget(rat, target, right_target, foundRight);
	end
	
	if ( not foundLeft ) then left_target.TrackingRat = ""; end
	if ( not foundRight ) then right_target.TrackingRat = ""; end
	
	scan_rats = false;
end

function updateMoraleValues(target)
	if ( target.TrackingRat ~= "" ) then
		local t = rat_targets[target.TrackingRat];
		target.CurrentMorale = t:GetMorale();
		target.MaximumMorale = t:GetMaxMorale();
	end
	target.PercentageLabel:SetText(getPercentage(target));
end

function updateDifferenceBar()
	diff_bar:SetWidth(math.floor(145*math.min(math.abs(left_target.CurrentMorale - right_target.CurrentMorale), MORALE_DIFFERENCE)/MORALE_DIFFERENCE));
	if left_target.CurrentMorale < right_target.CurrentMorale then
		diff_bar:SetLeft(152);
	else
		diff_bar:SetLeft(148 - diff_bar:GetWidth());
	end
	
end

function attachTargetListeners()
	rat_targets = {};
	rat_targets[my_player:GetName()] = my_player:GetTarget();
	my_player.TargetChanged = targetChanged;
	
	rat_targets_count = 1;
	local party = my_player:GetParty();
	if ( party ~= nil ) then
		for idx=1,party:GetAssistTargetCount() do
			local rat = party:GetAssistTarget(idx);
			rat_targets[rat:GetName()] = rat:GetTarget();
			rat.TargetChanged = targetChanged;
			rat_targets_count = rat_targets_count + 1;
		end
	end
	
	main_window.targetCounter:SetText(rat_targets_count);
end

-- Global vars
rat_targets = {};
rat_targets_count = 0;
my_player = Turbine.Gameplay.LocalPlayer.GetInstance();
main_window = nil;
settings_window = nil;
context_menu = nil;
diff_bar = nil;
scan_rats = true;
left_target = createTarget();
right_target = createTarget();

-- Global consts
UPDATE_FRAMECOUNT = 15;
MORALE_DIFFERENCE = 10000;

-- Load plugin
createMainWindow();
createSettingsWindow();
diff_bar = createDifferenceBar();
loadState();
Plugins["MoraleDiff"].Unload = saveState;
my_player.PartyChanged = reloadPlugin;
AddCallback(Turbine.Chat, "Received", processChatMessage);
AddCallback(main_window, "Update", onUpdate);
attachTargetListeners();
	
-- Run
main_window:SetVisible(true);
main_window:SetWantsUpdates(true);
